/***************************************************************************
                          functions.cpp  -  description
                             -------------------
    begin                : Feb 25, 2015
    copyright            : (C) 2015 by Tomáš Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <CPUCyclesCounter.h>

using namespace std;

int data[ 4 ];

class adder
{
   public:

      virtual void addNumber( int& n ) const = 0;
};

class oneAdder : public adder
{
   public:

      void addNumber( int& n ) const { n += data[ 0 ]; data[ 0 ] *= -1; };
};

template< int index >
class templateAdder
{
   public:
      void addNumber( int& n ) const { n += data[ index ]; data[ index ] *= -1; };

};

class switchAdder
{
   public:

      switchAdder( int c )
      : number( c ){};

      void addNumber( int& n ) const
      {
         switch( this->number )
         {
             case 0:
                n += data[ 0 ];
                break;
             case 1:
                n += data[ 1 ];
                break;
             case 2:
                n += data[ 2 ];
                break;
             case 3:
                n += data[ 3 ];
                break;
         }
         data[ this->number ] *= -1;
      }

   protected:

      int number;
};

class inlineAdder
{
   public:

      inline void addNumber( int& n ) const { n += data[ 0 ]; data[ 0 ] *= -1; };
};


int main( int argc, char* argv[] )
{
   CPUCyclesCounter cpuCycles;
   for( int i = 0; i < 4; i++ )
      data[ i ] = i+1;
   int sum( 0 );

   const unsigned  long long int loops = 1 << 15;

   oneAdder one_adder;
   cpuCycles.reset();
   cpuCycles.start();
   for( unsigned long long int j = 0; j < loops; j++)
      for( unsigned long long int i = 0; i < loops; i++ )
         one_adder.addNumber( sum );
   cpuCycles.stop();
   cout << "Method call " << ( double ) cpuCycles.getCycles() / ( double ) ( loops ) / ( double ) loops << " sum = " << sum <<  endl;

   adder* _adder = &one_adder;
   sum = 0;
   cpuCycles.reset();
   cpuCycles.start();
   for( unsigned long long int j = 0; j < loops; j++)
      for( unsigned long long int i = 0; i < loops; i++ )
         _adder->addNumber( sum );
   cpuCycles.stop();
   cout << "Virtual method call " << ( double ) cpuCycles.getCycles() / ( double ) ( loops ) / ( double ) loops << " sum = " << sum <<  endl;

   templateAdder< 1 > template_adder;
   sum = 0;
   cpuCycles.reset();
   cpuCycles.start();
   for( unsigned long long int j = 0; j < loops; j++)
      for( unsigned long long int i = 0; i < loops; i++ )
         template_adder.addNumber( sum );
   cpuCycles.stop();
   cout << "Template method call " << ( double ) cpuCycles.getCycles() / ( double ) ( loops ) / ( double ) loops << " sum = " << sum <<  endl;

   switchAdder switch_adder( 1 );
   sum = 0;
   cpuCycles.reset();
   cpuCycles.start();
   for( unsigned long long int j = 0; j < loops; j++)
      for( unsigned long long int i = 0; i < loops; i++ )
         switch_adder.addNumber( sum );
   cpuCycles.stop();
   cout << "Switch call " << ( double ) cpuCycles.getCycles() / ( double ) ( loops ) / ( double ) loops << " sum = " << sum <<  endl;

   sum = 0;
   cpuCycles.reset();
   cpuCycles.start();
   for( unsigned long long int j = 0; j < loops; j++)
      for( unsigned long long int i = 0; i < loops; i++ )
      {
         sum += data[ 0 ];
         data[ 0 ] *= -1;
      }
   cpuCycles.stop();
   cout << "No call " << ( double ) cpuCycles.getCycles() / ( double ) ( loops ) / ( double ) loops <<  " sum = " << sum << endl;

   inlineAdder inline_adder;
   sum = 0;
   cpuCycles.reset();
   cpuCycles.start();
   for( unsigned long long int j = 0; j < loops; j++)
      for( unsigned long long int i = 0; i < loops; i++ )
         inline_adder.addNumber( sum );
   cpuCycles.stop();
   cout << "Inline call " << ( double ) cpuCycles.getCycles() / ( double ) ( loops ) / ( double ) loops << " sum = " << sum <<  endl;

}
