/***************************************************************************
                          exceptions-test.cpp  -  description
                             -------------------
    begin                : 2015/02/04
    copyright            : (C) 2015 by Tomáš Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <CPUCyclesCounter.h>
#include <PapiCounter.h>

using namespace std;


void performTest( const int* data,
                  const int size,
                  int& a )
{
   a = 0;
   for( int i = 0; i < size; i++ )
   {
      if( data[ i ] == 1 )
         a++;
      if( data[ i ] == 0 )
         a = 0;
   }
}

void performTestWithException( const int* data,
                               const int size,
                               int& a )
{
   a = 0;
   for( int i = 0; i < size; i++ )
   {
      if( data[ i ] == 1 )
         a++;
      if( data[ i ] == 0 )
         throw -1;
   }
}

int main( int argc, char* argv[] )
{
   const int size = 256;
   const int loops = 2097152;

   int* data = new int[ size ];
   for( int i = 0; i < size; i++ )
      data[ i ] = 1;

   CPUCyclesCounter cpuCycles;
   int a;

   a = 0;
   cpuCycles.reset();
   cpuCycles.start();
   for( int i = 0; i < loops; i++ )
      performTest( data, size, a );
   cpuCycles.stop();
   cout << "No exceptions: " << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size ) << " ticks per element. Result is " << a << endl;

   a = 0;
   cpuCycles.reset();
   cpuCycles.start();
   for( int i = 0; i < loops; i++ )
   {
      try
      {
         performTestWithException( data, size, a );
      }
      catch( int e )
      {
         cerr << "Exception" << e << endl;
      }
   }
   cpuCycles.stop();
   cout << "Exceptions: " << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size ) << " ticks per element. Result is " << a << endl;

   data[ 0 ] = 0;
   a = 0;
   int exceptionsNum( 0 );
   cpuCycles.reset();
   cpuCycles.start();
   for( int i = 0; i < loops; i++ )
   {
      try
      {
         performTestWithException( data, size, a );
      }
      catch( int e )
      {
         exceptionsNum++;
      }
   }
   cpuCycles.stop();
   cout << "Exceptions throw: " << ( double ) cpuCycles.getCycles() / ( double ) ( loops ) << " ticks per element. Result is " << a
        << " number of exceptions is " << exceptionsNum++ << endl;

   delete[] data;
   return EXIT_SUCCESS;
}
