/***************************************************************************
                          MatrixAdditon.cpp  -  description
                             -------------------
    begin                : 2015/02/04
    copyright            : (C) 2015 by Tomáš Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstdlib>
#include <cstring>
#include <cmath>
#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <TimerRT.h>

using namespace std;

template< typename Real >
class ArrayMatrix
{
   public:
      ArrayMatrix( int size )
      {
         this->matrix = new Real[ size * size ];
         this->size = size;
      }

      void setValue( const Real& v )
      {
         for( int i = 0; i < size * size; i++ )
            this->matrix[ i ] = v;
      }

      void setElement( int i, int j, const Real& v )
      {
         this->matrix[ i * size + j ] = v;
      }

      ArrayMatrix& operator = ( const ArrayMatrix& m )
      {
         memcpy( this->matrix, m.matrix, size * size * sizeof( Real ) );
         return ( *this );
      }

      bool operator == ( const ArrayMatrix& m ) const
      {
         if( memcmp( this->matrix, m.matrix, size * size * sizeof( Real ) ) == 0 )
            return true;
         return false;
      }

      bool operator != ( const ArrayMatrix& m ) const
      {
         return ! this->operator == ( m );
      }

      void multiplyMatrices( const ArrayMatrix< Real >& m1,
                             const ArrayMatrix< Real >& m2 )
      {
         for( int i = 0; i < size; i++ )
            for( int j = 0; j < size; j++ )
            {
               Real aux( 0.0 );
               for( int k = 0; k < size; k++ )
                  aux += m1.matrix[ i * size + k ] * m2.matrix[ k * size + j ];
               this->matrix[ i * size + j ] = aux;
            }
      }

      void multiplyMatricesWithTransposition( const ArrayMatrix< Real >& m1,
                                              const ArrayMatrix< Real >& m2 )
      {
         Real* transposedMatrix = new Real[ size * size ];
         for( int i = 0; i < size; i++ )
            for( int j = 0; j < size; j++ )
               transposedMatrix[ i * size + j] =
                  m2.matrix[ j * size + i ];

         for( int i = 0; i < size; i++ )
            for( int j = 0; j < size; j++ )
            {
               Real aux( 0.0 );
               for( int k = 0; k < size; k++ )
                  aux += m1.matrix[ i * size + k ] * transposedMatrix[ j * size + k ];
               this->matrix[ i * size + j ] = aux;
            }

         delete[] transposedMatrix;
      }

      template< int blockSize = 16 >
      void multiplyMatricesBlockwise1( const ArrayMatrix< Real >& m1,
                                       const ArrayMatrix< Real >& m2 )
      {
         memset( this->matrix, 0, size * size * sizeof( Real  ) );
         for( int i = 0; i < size; i += blockSize )
            for( int j = 0; j < size; j += blockSize )
               for( int k = 0; k < size; k += blockSize )
               {
                  for( int i1 = i; i1 < i + blockSize; i1++ )
                     for( int j1 = j; j1 < j + blockSize; j1++ )
                        for( int k1 = k; k1 < k + blockSize; k1++ )
                           this->matrix[ i1*size + j1 ] +=
                              m1.matrix[ i1*size + k1 ]*
                                 m2.matrix[ k1*size  + j1 ];
               }
      }
      template< int blockSize = 16 >
      void multiplyMatricesBlockwise2( const ArrayMatrix< Real >& m1,
                                       const ArrayMatrix< Real >& m2 )
      {
         memset( this->matrix, 0, size * size * sizeof( Real  ) );
         for( int i = 0; i < size; i += blockSize )
            for( int j = 0; j < size; j += blockSize )
               for( int k = 0; k < size; k += blockSize )
               {
                  Real* res = &this->matrix[ i * size + j ];
                  Real* _m1 = &m1.matrix[ i*size + k ];
                  for( int i1 = 0; i1 < blockSize; i1++, res += size, _m1 += size )
                     for( int j1 = 0; j1 < blockSize; j1++ )
                     {
                        Real* _m2 = &m2.matrix[ k*size + j ];
                        for( int k1 = 0; k1 < blockSize; k1++, _m2 += size )
                           res[ j1 ] += _m1[ k1 ] * _m2[ j1 ];
                     }
               }
      }

      template< int blockSize = 16 >
      void multiplyMatricesBlockwise3( const ArrayMatrix< Real >& m1,
                                       const ArrayMatrix< Real >& m2 )
      {
         memset( this->matrix, 0, size * size * sizeof( Real  ) );
         for( int i = 0; i < size; i += blockSize )
           for( int j = 0; j < size; j += blockSize )
              for( int k = 0; k < size; k += blockSize )
              {
                 Real* res = &this->matrix[ i * size + j ];
                 Real* _m1 = &m1.matrix[ i*size + k ];
                 for( int i1 = 0; i1 < blockSize; i1++, res += size, _m1 += size )
                 {
                    Real* _m2 = &m2.matrix[ k*size + j ];
                    for( int k1 = 0; k1 < blockSize; k1++, _m2 += size )
                       for( int j1 = 0; j1 < blockSize; j1++ )
                          res[ j1 ] += _m1[ k1 ] * _m2[ j1 ];
                 }
              }
      }

      ~ArrayMatrix()
      {
         delete[] this->matrix;
      }
   protected:
      Real* matrix;

      int size;
};

void writeTableHeader( ostream& str )
{
   str << std::right;
   str << "#";
   str << setw( 195 ) << std::setfill( '-') << "-" << endl;
   str << "#";
   str << std::setfill( ' ' );
   str << setw( 30 ) << "Matrix size       "
       << setw( 30 ) << "Common"
       << setw( 30 ) << "Transposition"
       << setw( 30 ) << "Blocks 1"
       << setw( 30 ) << "Blocks 2"
       << setw( 30 ) << "Blocks 3"
       << setw( 15 ) << "Speed-up"
       << endl;
   str << "#";
   str << setw( 15 ) << "size"
       << setw( 15 ) << "log2"
       << setw( 15 ) << "s"
       << setw( 15 ) << "GFLOPS"
       << setw( 15 ) << "s"
       << setw( 15 ) << "GFLOPS"
       << setw( 15 ) << "s"
       << setw( 15 ) << "GFLOPS"
       << setw( 15 ) << "s"
       << setw( 15 ) << "GFLOPS"
       << setw( 15 ) << "s"
       << setw( 15 ) << "GFLOPS"
       << endl;
   str << "#";
   str << setw( 195 ) << std::setfill( '-') << "-" << endl;
   str << std::setfill( ' ' );
}

void writeTableLine( ostream& str,
                     const int size,
                     const double& commonMultiplicationTime,
                     const double& commonMultiplicationGflops,
                     const double& transposeMultiplicationTime,
                     const double& transposeMultiplicationGflops,
                     const double& blockwise1MultiplicationTime,
                     const double& blockwise1MultiplicationGflops,
                     const double& blockwise2MultiplicationTime,
                     const double& blockwise2MultiplicationGflops,
                     const double& blockwise3MultiplicationTime,
                     const double& blockwise3MultiplicationGflops )
{
   str << setw( 15 ) << size
       << setw( 15 ) << log2( size )
       << setw( 15 ) << commonMultiplicationTime
       << setw( 15 ) << commonMultiplicationGflops
       << setw( 15 ) << transposeMultiplicationTime
       << setw( 15 ) << transposeMultiplicationGflops
       << setw( 15 ) << blockwise1MultiplicationTime
       << setw( 15 ) << blockwise1MultiplicationGflops
       << setw( 15 ) << blockwise2MultiplicationTime
       << setw( 15 ) << blockwise2MultiplicationGflops
       << setw( 15 ) << blockwise3MultiplicationTime
       << setw( 15 ) << blockwise3MultiplicationGflops
       << setw( 15 ) << commonMultiplicationTime / blockwise3MultiplicationTime
       << endl;
}

int main( int argc, char* argv[] )
{
   const long int maxSize = 8192;
   const double minTestTime( 1.0 );
   TimerRT timer;

   fstream outputFile;
   outputFile.open( "results.txt", ios::out );
   if( ! outputFile )
   {
      cerr << "I cannot open the output file results.txt." << endl;
      return EXIT_FAILURE;
   }

   writeTableHeader( cout );
   writeTableHeader( outputFile );
   long long int size = 16;
   while( size < maxSize )
   {
      const double operations = 2 * size * size * size;
      ArrayMatrix< float > a( size ), b( size ), c( size ), d( size );
      a.setValue( 1.0 );
      b.setValue( 0.0 );
      for( int i = 0; i < size; i++ )
         b.setElement( i, i, i + 1 );

      int counter;

      counter = 0;
      timer.reset();
      timer.start();
      while( timer.getTime() < minTestTime )
      {
         c.multiplyMatrices( a, b );
         counter++;
      }
      timer.stop();
      double slowTime = timer.getTime() / ( double ) counter;
      double slowMultiplyGflops = operations / slowTime / 1.0e9;
      d = c;

      counter = 0;
      timer.reset();
      timer.start();
      while( timer.getTime() < minTestTime )
      {
         c.multiplyMatricesWithTransposition( a, b );
         counter++;
      }
      timer.stop();
      double transposeTime = timer.getTime() / ( double ) counter;
      double transposeMultiplyGflops = operations / transposeTime / 1.0e9;
      if( d != c )
      {
         cerr << "Multiplication with transpostion error." << endl;
         return EXIT_FAILURE;
      }

      counter = 0;
      timer.reset();
      timer.start();
      while( timer.getTime() < minTestTime )
      {
         c.template multiplyMatricesBlockwise1< 16 >( a, b );
         counter++;
      }
      timer.stop();
      double blockwise1Time = timer.getTime() / ( double ) counter;
      double blockwise1MultiplyGflops = operations / blockwise1Time / 1.0e9;
      if( d != c )
      {
         cerr << "Multiplication with blocks v.1. error." << endl;
         return EXIT_FAILURE;
      }

      counter = 0;
      timer.reset();
      timer.start();
      while( timer.getTime() < minTestTime )
      {
         c.template multiplyMatricesBlockwise2< 16 >( a, b );
         counter++;
      }
      timer.stop();
      double blockwise2Time = timer.getTime() / ( double ) counter;
      double blockwise2MultiplyGflops = operations / blockwise2Time / 1.0e9;
      if( d != c )
      {
         cerr << "Multiplication with blocks v.2. error." << endl;
         return EXIT_FAILURE;
      }

      counter = 0;
      timer.reset();
      timer.start();
      while( timer.getTime() < minTestTime )
      {
         c.template multiplyMatricesBlockwise3< 16 >( a, b );
         counter++;
      }
      timer.stop();
      double blockwise3Time = timer.getTime() / ( double ) counter;
      double blockwise3MultiplyGflops = operations / blockwise3Time / 1.0e9;
      if( d != c )
      {
         cerr << "Multiplication with blocks v.3. error." << endl;
         return EXIT_FAILURE;
      }



      writeTableLine( cout,
                      size,
                      slowTime,
                      slowMultiplyGflops,
                      transposeTime,
                      transposeMultiplyGflops,
                      blockwise1Time,
                      blockwise1MultiplyGflops,
                      blockwise2Time,
                      blockwise2MultiplyGflops,
                      blockwise3Time,
                      blockwise3MultiplyGflops );
      writeTableLine( outputFile,
                      size,
                      slowTime,
                      slowMultiplyGflops,
                      transposeTime,
                      transposeMultiplyGflops,
                      blockwise1Time,
                      blockwise1MultiplyGflops,
                      blockwise2Time,
                      blockwise2MultiplyGflops,
                      blockwise3Time,
                      blockwise3MultiplyGflops );
      size *= 2;
   }
}

