/***************************************************************************
                          memory-allocation.cpp  -  description
                             -------------------
    begin                : Feb 23, 2015
    copyright            : (C) 2015 by Tomá¹ Oberhuber
    email                : oberhuber@seznam.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <alloca.h>
#include <stdlib.h>
#include <CPUCyclesCounter.h>

using namespace std;

int a[ 512 ];

void globalArrayTest( int pos )
{
   a[ pos - 1 ] = 1;
}

void localArrayTest( int pos )
{
   int b[ 512 ];
   b[ pos - 1 ] = 1;
}

void stackTest( int pos )
{
   int* c = (int* )  alloca( pos * sizeof( int ) );
   c[ pos - 1 ] = 1;
}

void heapTest( int pos )
{
   int* c = (int* ) malloc( pos * sizeof( int ) );
   c[ pos - 1 ] = 1;
   free( ( void* ) c );
}

int main( int argc, char* argv[] )
{
   const int testLoops = 1<<24;

   CPUCyclesCounter cpuCycles;

   cpuCycles.reset();
   cpuCycles.start();
   for( int i = 0; i < testLoops; i++ )
      globalArrayTest( 512 );
   cpuCycles.stop();
   cout << "The global array test took: " << ( double ) cpuCycles.getCycles() / ( double ) testLoops << " CPU cycles." << endl;


   cpuCycles.reset();
   cpuCycles.start();
   for( int i = 0; i < testLoops; i++ )
      localArrayTest( 512 );
   cpuCycles.stop();
   cout << "The local array test took: " << ( double ) cpuCycles.getCycles() / ( double ) testLoops << " CPU cycles." << endl;


   cpuCycles.reset();
   cpuCycles.start();
   for( int i = 0; i < testLoops; i++ )
      stackTest( 512 );
   cpuCycles.stop();
   cout << "The stack test took: " << ( double ) cpuCycles.getCycles() / ( double ) testLoops << " CPU cycles." << endl;

   cpuCycles.reset();
   cpuCycles.start();
   for( int i = 0; i < testLoops; i++ )
      heapTest( 512 );
   cpuCycles.stop();
   cout << "The heap test took: " << ( double ) cpuCycles.getCycles() / ( double ) testLoops << " CPU cycles." << endl;

}
