/***************************************************************************
                          operations-test.cpp  -  description
                             -------------------
    begin                : Mar 1, 2015
    copyright            : (C) 2015 by Tomas Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <math.h>
#include <stdlib.h>
#include <CPUCyclesCounter.h>

using namespace std;

template< typename NumType >
void computeSum( const NumType& n1,
                 const NumType& n2,
                 const NumType& n3,
                 const NumType& n4,
                 const NumType& n5,
                 const NumType& n6,
                 const NumType& n7,
                 const NumType& n8,
                 NumType& sum)
{
   sum += n1 + n2 + n3 + n4 + n5 + n6 + n7 + n8;
}

template< typename NumType >
void computeSumFast( const NumType& n1,
                     const NumType& n2,
                     const NumType& n3,
                     const NumType& n4,
                     const NumType& n5,
                     const NumType& n6,
                     const NumType& n7,
                     const NumType& n8,
                     NumType& sum)
{
   sum += ( n1 + n2 ) + ( n3 + n4 ) + ( n5 + n6 ) + ( n7 + n8 );
}

template< typename NumType >
void computeSumFast2( const NumType& n1,
                      const NumType& n2,
                      const NumType& n3,
                      const NumType& n4,
                      const NumType& n5,
                      const NumType& n6,
                      const NumType& n7,
                      const NumType& n8,
                      NumType& sum)
{
   sum += ( ( n1 + n2 ) + ( n3 + n4 ) ) + ( ( n5 + n6 ) + ( n7 + n8 ) );
}

int main( int argc, char* argv[] )
{
   CPUCyclesCounter cpuCycles;
   const int size = 65536;
   int* integerData = new int[ size ];
   float* floatData = new float[ size ];
   double* doubleData = new double[ size ];

   for( int i = 0; i < size; i++ )
      integerData[ i ] = floatData[ i ] = doubleData[ i ] = 9;

   const unsigned  long long int loops = 1024;
   int integerResult( 0 );
   float floatResult( 0.0 );
   double doubleResult( 0.0 );

   cpuCycles.reset();
   cpuCycles.start();
   integerResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size - 8; i++ )
         computeSum( integerData[ i ],
                     integerData[ i + 1 ],
                     integerData[ i + 2 ],
                     integerData[ i + 3 ],
                     integerData[ i + 4 ],
                     integerData[ i + 5 ],
                     integerData[ i + 6 ],
                     integerData[ i + 7 ],
                     integerResult );
   cpuCycles.stop();
   cout << setw( 30 ) << "Integer sum: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << integerResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   integerResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size - 8; i++ )
         computeSumFast( integerData[ i ],
                         integerData[ i + 1 ],
                         integerData[ i + 2 ],
                         integerData[ i + 3 ],
                         integerData[ i + 4 ],
                         integerData[ i + 5 ],
                         integerData[ i + 6 ],
                         integerData[ i + 7 ],
                         integerResult );
   cpuCycles.stop();
   cout << setw( 30 ) << "Integer fast sum: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << integerResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size - 8; i++ )
         computeSum( floatData[ i ],
                     floatData[ i + 1 ],
                     floatData[ i + 2 ],
                     floatData[ i + 3 ],
                     floatData[ i + 4 ],
                     floatData[ i + 5 ],
                     floatData[ i + 6 ],
                     floatData[ i + 7 ],
                     floatResult );
   cpuCycles.stop();
   cout << setw( 30 ) << "Float sum: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size - 8; i++ )
         computeSumFast( floatData[ i ],
                         floatData[ i + 1 ],
                         floatData[ i + 2 ],
                         floatData[ i + 3 ],
                         floatData[ i + 4 ],
                         floatData[ i + 5 ],
                         floatData[ i + 6 ],
                         floatData[ i + 7 ],
                         floatResult );

   cpuCycles.stop();
   cout << setw( 30 ) << "Float fast sum: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size - 8; i++ )
         computeSumFast2( floatData[ i ],
                          floatData[ i + 1 ],
                          floatData[ i + 2 ],
                          floatData[ i + 3 ],
                          floatData[ i + 4 ],
                          floatData[ i + 5 ],
                          floatData[ i + 6 ],
                          floatData[ i + 7 ],
                          floatResult );
   cpuCycles.stop();
   cout << setw( 30 ) << "Float fast sum 2: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size - 8; i++ )
         computeSum( doubleData[ i ],
                     doubleData[ i + 1 ],
                     doubleData[ i + 2 ],
                     doubleData[ i + 3 ],
                     doubleData[ i + 4 ],
                     doubleData[ i + 5 ],
                     doubleData[ i + 6 ],
                     doubleData[ i + 7 ],
                     doubleResult );
   cpuCycles.stop();
   cout << setw( 30 ) << "Double sum: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size - 8; i++ )
         computeSumFast( doubleData[ i ],
                         doubleData[ i + 1 ],
                         doubleData[ i + 2 ],
                         doubleData[ i + 3 ],
                         doubleData[ i + 4 ],
                         doubleData[ i + 5 ],
                         doubleData[ i + 6 ],
                         doubleData[ i + 7 ],
                         doubleResult );
   cpuCycles.stop();
   cout << setw( 30 ) << "Double fast sum: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size - 8; i++ )
         computeSumFast2( doubleData[ i ],
                          doubleData[ i + 1 ],
                          doubleData[ i + 2 ],
                          doubleData[ i + 3 ],
                          doubleData[ i + 4 ],
                          doubleData[ i + 5 ],
                          doubleData[ i + 6 ],
                          doubleData[ i + 7 ],
                          doubleResult );
   cpuCycles.stop();
   cout << setw( 30 ) << "Double fast sum 2: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   integerResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         integerResult += integerData[ i ] % 3;
   cpuCycles.stop();
   cout << setw( 30 ) << "Integer mod 3: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << integerResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   integerResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         integerResult += integerData[ i ] % 2;
   cpuCycles.stop();
   cout << setw( 30 ) << "Integer mod 2: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << integerResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   integerResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         integerResult += integerData[ i ] / 3;
   cpuCycles.stop();
   cout << setw( 30 ) << "Integer division: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << integerResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   integerResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         integerResult += integerData[ i ] * 3;
   cpuCycles.stop();
   cout << setw( 30 ) << "Integer multiplication: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << integerResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         floatResult += floatData[ i ] / 3;
   cpuCycles.stop();
   cout << setw( 30 ) << "Float division: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         doubleResult += doubleData[ i ] / 3;
   cpuCycles.stop();
   cout << setw( 30 ) << "Double division: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         floatResult += floatData[ i ] * ( float ) ( 1.0 / 3.0 );
   cpuCycles.stop();
   cout << setw( 30 ) << "Float multiplication: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;


   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         doubleResult += doubleData[ i ] * ( 1.0 / 3.0 );
   cpuCycles.stop();
   cout << setw( 30 ) << "Double multiplication: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         floatResult += pow( floatData[ i ], ( float ) 2.0 );
   cpuCycles.stop();
   cout << setw( 30 ) << "Float pow 2: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         doubleResult += pow( doubleData[ i ], 2.0 );
   cpuCycles.stop();
   cout << setw( 30 ) << "Double pow 2: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         floatResult += pow( floatData[ i ], ( float ) 2.13 );
   cpuCycles.stop();
   cout << setw( 30 ) << "Float pow 2.13: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         doubleResult += pow( doubleData[ i ], 2.13 );
   cpuCycles.stop();
   cout << setw( 30 ) << "Double pow 2.13: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         floatResult += sqrt( floatData[ i ] );
   cpuCycles.stop();
   cout << setw( 30 ) << "Float sqrt: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         doubleResult += sqrt( doubleData[ i ] );
   cpuCycles.stop();
   cout << setw( 30 ) << "Double sqrt: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         floatResult += exp( floatData[ i ] );
   cpuCycles.stop();
   cout << setw( 30 ) << "Float exp: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         doubleResult += exp( doubleData[ i ] );
   cpuCycles.stop();
   cout << setw( 30 ) << "Double exp: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         floatResult += log10( floatData[ i ] );
   cpuCycles.stop();
   cout << setw( 30 ) << "Float log: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         doubleResult += log10( doubleData[ i ] );
   cpuCycles.stop();
   cout << setw( 30 ) << "Double log: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;


   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         floatResult += sin( floatData[ i ] );
   cpuCycles.stop();
   cout << setw( 30 ) << "Float sin: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         doubleResult += sin( doubleData[ i ] );
   cpuCycles.stop();
   cout << setw( 30 ) << "Double sin: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleResult <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   floatResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         floatData[ i ] = integerData[ i ];
   cpuCycles.stop();
   cout << setw( 30 ) << "int -> float: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << floatData[ size / 2 ] <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         doubleData[ i ] = integerData[ i ];
   cpuCycles.stop();
   cout << setw( 30 ) << "int -> double: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleData[ size / 2 ] <<  endl;

   cpuCycles.reset();
   cpuCycles.start();
   doubleResult = 0;
   for( unsigned long long int j = 0; j < loops; j++)
      for( int i = 0; i < size; i++ )
         doubleData[ i ] = floatData[ i ];
   cpuCycles.stop();
   cout << setw( 30 ) << "float -> double: "
        << setw( 15 ) << ( double ) cpuCycles.getCycles() / ( double ) ( loops * size )
        << setw( 25 ) << " CPU ticks/ element. result = " << doubleData[ size / 2 ] <<  endl;

   delete[] integerData;
   delete[] floatData;
   delete[] doubleData;
}
